unit MainStorageUnit;

interface

uses
  Windows, Classes, ActiveX, ShellApi, SysUtils, Dialogs, ComObj, SxTools,
  AxCtrls, Graphics, SxIconList;

type
  TStgFolder=class;
  TStgFile=class;

  TMainStorage=class
  private
    FStorage: IStorage;
    FNamespace:IStorage;
    FFileName: String;
    procedure SetFilename(const Value: String);
    function GetNamespace: IStorage;
  protected
    function OpenStorage:Boolean;
    procedure CloseStorage;
    procedure ReadSettings;
    procedure WriteSettings;
    property Storage:IStorage read FStorage write FStorage;
    property Namespace:IStorage read GetNamespace;
  public
    constructor Create;
    destructor Destroy;override;
    function CreateStorage(AFileName:String):Boolean;
    function GetElementStorage(stgPath:String;CanCreate:Boolean):IStorage;
    function OpenFolder(FolderPath:String):TStgFolder;
    function OpenFile(FilePath:String):TStgFile;
    property FileName:String read FFileName write SetFilename;
  end;

  TStgAttributes=record
    ReadOnly:Boolean;
  end;

  TStgObject=class
  private
    FName: String;
    FOwner: TMainStorage;
    FAttributes: TStgAttributes;
    FPath: String;
    FElementStorage: IStorage;
    FIcon: TIcon;
    FOpenIcon:TIcon;
    FIsSystemIcon: Boolean;
    function GetIcon: TIcon;
    function GetOpenIcon: TIcon;
    procedure SetIcon(const Value: TIcon);
    procedure SetOpenIcon(const Value: TIcon);
  public
    constructor Create(AOwner:TMainStorage;APath:String);virtual;
    destructor Destroy;override;
    procedure ReadElementAttributes;virtual;
    procedure WriteElementAttributes;virtual;
    property Owner:TMainStorage read FOwner write FOwner;
    property Name:String read FName write FName;
    property Path:String read FPath write FPath;
    property Icon:TIcon read GetIcon write SetIcon;
    property OpenIcon:TIcon read GetOpenIcon write SetOpenIcon;
    property IsSystemIcon:Boolean read FIsSystemIcon;
    property Attributes:TStgAttributes read FAttributes write FAttributes;
    property ElementStorage:IStorage read FElementStorage write FElementStorage;
  end;

  TStgFolder=class(TStgObject)
  public
    constructor Create(AOwner:TMainStorage;APath:String);override;
  end;

  TStgFile=class(TStgObject)
  public
    constructor Create(AOwner:TMainStorage;APath:String);override;
  end;

const
  stgMode=STGM_READWRITE or STGM_SHARE_DENY_NONE or STGM_DIRECT;
  nameAttributes='!_Attri_butes_!';
  nameNamespace='Namespace';

var
  MainStorage:TMainStorage;

implementation

uses
  Registry;

{ TMainStorage }

procedure TMainStorage.CloseStorage;
begin
  Storage:=nil;
end;

constructor TMainStorage.Create;
begin
  FStorage:=nil;
  FNamespace:=nil;
  ReadSettings;
end;

function TMainStorage.CreateStorage(AFilename:String): Boolean;
var
  tempFilename:WideString;
  tempStorage:IStorage;
begin
  try
    tempFileName:=AFilename;
    OleCheck(StgCreateDocfile(PWideChar(tempFileName),STGM_READWRITE or STGM_DIRECT or STGM_SHARE_EXCLUSIVE, 0,tempStorage));
    Storage:=tempStorage;
    FFileName:=AFilename;
    Result:=True;
  except
    Result:=False;
  end;
end;

destructor TMainStorage.Destroy;
begin
  CloseStorage;
  inherited;
end;

function TMainStorage.GetElementStorage(stgPath: String;
  CanCreate: Boolean): IStorage;
var
  tempStr:TStrings;
  tempResult:IStorage;
  tempName:WideString;
  tres:IStorage;
  i:Integer;
begin
  Result:=nil;
  tempStr:=TStringList.Create;
  try
    FilePathToStrings(stgPath,tempStr);
    tempResult:=Namespace;
    for i:=0 to tempStr.Count-1 do
    begin
      tres:=nil;
      tempName:=tempStr[i];
      if Failed(tempResult.OpenStorage(PWideChar(tempName),nil,stgMode,nil,0,tres))
         and CanCreate then
         if Failed(tempResult.CreateStorage(PWideChar(tempName),stgMode,0,0,tres)) then
            tres:=nil;
      tempResult:=tres;
      if tempResult=nil then
         Break;
    end;
    Result:=tempResult;
  finally
    FreeAndNil(tempStr);
  end;
end;

function TMainStorage.GetNamespace: IStorage;
begin
  try
    if FNamespace=nil then
    begin
      if Failed(Storage.OpenStorage(nameNamespace,nil,stgMode,nil,0,FNamespace)) then
         if Failed(Storage.CreateStorage(nameNamespace,stgMode,0,0,FNamespace)) then
            raise Exception.Create('Can''t create storage for namespace');
    end;
    Result:=FNamespace
  except
    Result:=nil;
  end;
end;

function TMainStorage.OpenFile(FilePath: String): TStgFile;
begin
  Result:=nil;
end;

function TMainStorage.OpenFolder(FolderPath: String): TStgFolder;
begin
  Result:=nil;
end;

function TMainStorage.OpenStorage:Boolean;
var
  tempFilename:WideString;
  tempStorage:IStorage;
begin
  try
    tempFileName:=Filename;
    Result:=Succeeded(StgOpenStorage(PWideChar(tempFileName),nil,stgMode,nil,0,tempStorage));
    if not Result then
       Exit;
    Storage:=tempStorage;
    Result:=True;
  except
    Result:=False;
  end;
end;

procedure TMainStorage.ReadSettings;
var
  Reg:TRegistry;
begin
  Reg:=TRegistry.Create;
  try
    Reg.OpenKey('\Software\ShellPlus\Examples\MegaDemoNSE',True);
    if Reg.ValueExists('StorageFile') then
       FileName:=Reg.ReadString('StorageFile');
  finally
    FreeAndNil(Reg);
  end;
end;

procedure TMainStorage.SetFilename(const Value: String);
begin
  if Value<>Filename then
  begin
    CloseStorage;
    FFileName := Value;
    if OpenStorage then
       WriteSettings;
  end;
end;

procedure TMainStorage.WriteSettings;
var
  Reg:TRegistry;
begin
  Reg:=TRegistry.Create;
  try
    Reg.OpenKey('\Software\ShellPlus\Examples\MegaDemoNSE',True);
    Reg.WriteString('StorageFile',Filename);
  finally
    FreeAndNil(Reg);
  end;
end;

{ TStgObject }

constructor TStgObject.Create(AOwner: TMainStorage;APath:String);
var
  tempStr:TStrings;
begin
  FOwner:=AOwner;
  FIsSystemIcon:=True;
  FIcon:=TIcon.Create;
  FOpenIcon:=TIcon.Create;
  tempStr:=TStringList.Create;
  try
    FilePathToStrings(APath,tempStr);
    if tempStr.Count=0 then
       raise Exception.Create('Can''t create element without name');
    FName:=tempStr[tempStr.Count-1]; // storing element name
    tempStr.Delete(tempStr.Count-1); // removing last element to store only path
    FPath:=StringsToFilePath(tempStr);
    FElementStorage:=Owner.GetElementStorage(APath, True);
    if ElementStorage=nil then
       raise Exception.Create('Can''t create storage for element');
    ReadElementAttributes;
  finally
    FreeAndNil(tempStr);
  end;
end;

destructor TStgObject.Destroy;
begin
  FreeAndNil(FIcon);
  FreeAndNil(FOpenIcon);
  inherited;
end;

function TStgObject.GetIcon: TIcon;
begin
  if FIcon.Empty then
     FIcon.Handle:=SxGetFileSystemIconHandle(Name,True,False,Self is TStgFolder);
  Result := FIcon;
end;

function TStgObject.GetOpenIcon: TIcon;
begin
  if FOpenIcon.Empty then
     FOpenIcon.Handle:=SxGetFileSystemIconHandle(Name,True,True,Self is TStgFolder);
  Result := FIcon;
end;

procedure TStgObject.ReadElementAttributes;
var
  tempAttr:IStream;
  tempStream:TOleStream;
  SxStream:TSxStream;
  tempIconList:TSxIconList;
begin
  if Failed(ElementStorage.OpenStream(nameAttributes,nil,stgMode,0,tempAttr)) then
  begin
    if Failed(ElementStorage.CreateStream(nameAttributes,stgMode,0,0,tempAttr)) then
       raise Exception.Create('Failed to create attributes storage for element');
    Exit;
  end;
  tempStream:=TOleStream.Create(tempAttr);
  try
    SxStream:=TSxStream.Create(tempStream);
    try
      // Reading icon kinds
      FIsSystemIcon:=SxStream.ReadBoolean;
      // Reading icons if stored
      if not IsSystemIcon then
      begin
        tempIconList:=TSxIconList.Create(nil);
        try
          tempIconList.LoadFromStream(tempStream);
          if tempIconList.IconCount<2 then
             FIsSystemIcon:=True
          else
          begin
            FIcon.Handle:=tempIconList.GetHIcon(0);
            FOpenIcon.Handle:=tempIconList.GetHIcon(1)
          end;
        finally
          FreeAndNil(tempIconList);
        end;
      end;
      // Reading attributes
      FAttributes.ReadOnly:=SxStream.ReadBoolean;
    finally
      FreeAndNil(SxStream);
    end;
  finally
    FreeAndNil(tempStream);
  end;
end;

procedure TStgObject.SetIcon(const Value: TIcon);
begin
  FIsSystemIcon:=False;
  FIcon.Assign(Value);
end;

procedure TStgObject.SetOpenIcon(const Value: TIcon);
begin
  FIsSystemIcon:=False;
  FOpenIcon.Assign(Value);
end;

procedure TStgObject.WriteElementAttributes;
var
  tempAttr:IStream;
  tempStream:TOleStream;
  SxStream:TSxStream;
  tempIconList:TSxIconList;
begin
  if Failed(ElementStorage.OpenStream(nameAttributes,nil,stgMode,0,tempAttr)) then
  begin
    if Failed(ElementStorage.CreateStream(nameAttributes,stgMode,0,0,tempAttr)) then
       raise Exception.Create('Failed to create attributes storage for element');
  end
  else
    tempAttr.SetSize(0);
  tempStream:=TOleStream.Create(tempAttr);
  try
    SxStream:=TSxStream.Create(tempStream);
    try
      // Writing icon kinds
      SxStream.WriteBoolean(IsSystemIcon);
      // Writing icons if stored
      if not IsSystemIcon then
      begin
        tempIconList:=TSxIconList.Create(nil);
        try
          tempIconList.AddIcon(FIcon.Handle);
          tempIconList.AddIcon(FOpenIcon.Handle);
          tempIconList.SaveToStream(tempStream);
        finally
          FreeAndNil(tempIconList);
        end;
      end;
      // Writing attributes
      SxStream.WriteBoolean(Attributes.ReadOnly);
    finally
      FreeAndNil(SxStream);
    end;
  finally
    FreeAndNil(tempStream);
  end;
end;

{ TStgFolder }

constructor TStgFolder.Create(AOwner: TMainStorage;APath:String);
begin
  inherited;

end;

{ TStgFile }

constructor TStgFile.Create(AOwner: TMainStorage; APath: String);
begin
  inherited;

end;

initialization
  MainStorage:=TMainStorage.Create;

finalization
  FreeAndNil(MainStorage);

end.
